module ("L_VistaAlarmPanel1", package.seeall)

-- Flags
local DEBUG_MODE = false

-- Constants
local VERSION = "2.21"
local DEBUG_MODE_DEFAULT = "0"

local g_nPartitions -- Number of partitions.

local TTL = 60 -- If the panel doesn't report a zone as faulted in this number of seconds, the zone is considered unfaulted.
               -- The timer is reset every time a fault report for this zone is received.
local TIMER_INTERVAL = 5 -- The delay used in the 'ttlCountdownTimer' function (luup.call_delay).
local SLEEP_MS = 250

local SID = {
	PANEL = "urn:micasaverde-com:serviceId:VistaAlarmPanel1",
	PARTITION = "urn:micasaverde-com:serviceId:AlarmPartition2"
}

local TASK = {
	ERROR = 2,
	ERROR_PERM = -2,
	SUCCESS = 4,
	BUSY = 1
}

-- Byte 2: 07,06,05,04,03,02,01,00
-- Byte 3: 15,14,13,12,11,10,09,08
-- Byte 4: 23,22,21,20,19.18,17,16
-- Byte 5: 31,30,29,28,27,26,25,24
local ADD_IDX = { -- Address Index
	["00"] = 25, ["01"] = 26, ["02"] = 27, ["03"] = 28, ["04"] = 29, ["05"] = 30, ["06"] = 31, ["07"] = 32,
	["08"] = 17, ["09"] = 18, ["10"] = 19, ["11"] = 20, ["12"] = 21, ["13"] = 22, ["14"] = 23, ["15"] = 24,
	["16"] =  9, ["17"] = 10, ["18"] = 11, ["19"] = 12, ["20"] = 13, ["21"] = 14, ["22"] = 15, ["23"] = 16,
	["24"] =  1, ["25"] =  2, ["26"] =  3, ["27"] =  4, ["28"] =  5, ["29"] =  6, ["30"] =  7, ["31"] =  8
}

-- Globals
local lug_device = nil
local g_taskHandle = -1
local g_userCodes = {} -- Used when a lettered key is pressed.
local g_faultedZones = {} -- Stores the faulted zones list.

local g_addrPart = {} -- Table where the keys are the keypad addresses
                      -- and the values are the partitions assigned to the keypads.
local g_noAddrPart = true -- True if one or more partitions have no associated keypads.

local g_partitions = {
	-- device
	-- armMode
	-- detailedArmMode
	-- chimeMode
	-- pinCode
}


-------------------------------------------------------------------------------
--------------------------- Panel Related Constants ---------------------------
-------------------------------------------------------------------------------


local armingCommands = {
	["Disarmed"] = "1",     -- "DISARMED"
	["Armed"] = "2",        -- "AWAY"
	["ArmedInstant"] = "4", -- "MAXIMUM"
	["Stay"] = "3",         -- "STAY"
	["StayInstant"] = "7",  -- "INSTANT"
	["Night"] = "33",       -- "NIGHT-STAY"
	["NightInstant"] = "33",-- "NIGHT-STAY"
	["Force"] = "2",        -- "AWAY"
	["Vacation"] = "4"      -- "MAXIMUM"
}

local keyCommands = {
	["A"] = string.char (1):rep (3),
	["B"] = string.char (2):rep (3),
	["C"] = string.char (3):rep (3),
	["D"] = string.char (4):rep (3)
}


-------------------------------------------------------------------------------
------------------------------ Generic Utilities ------------------------------
-------------------------------------------------------------------------------


local log = luup.log

local function debug (text)
	if (DEBUG_MODE) then
		log (text)
	end
end


local function task (text, mode)
	debug (string.format ("(task) %s", text))
	if (mode == TASK.ERROR_PERM) then
		g_taskHandle = luup.task (text, TASK.ERROR, "Vista Alarm Panel", g_taskHandle)
	else
		g_taskHandle = luup.task (text, mode, "Vista Alarm Panel", g_taskHandle)

		-- Clear the previous error, since they're all transient
		if (mode ~= TASK.SUCCESS) then
			luup.call_delay ("clearTask", 30, "", false)
		end
	end
end


function clearTask()
	task ("Clearing...", TASK.SUCCESS)
	return true
end


function string.split (s, delimiter)
	if (not s or s == "") then
		return {}
	end

	local result = {}
	local from = 1
	local delimFrom, delimTo = s:find (delimiter, from)
	while delimFrom do
		table.insert (result, s:sub (from, delimFrom - 1))
		from  = delimTo + 1
		delimFrom, delimTo = s:find (delimiter, from)
	end
	table.insert (result, s:sub (from))

	return result
end


function table.contains (t, item)
	if (t) then
		for k, v in pairs (t) do
			if (v == item) then
				return true, k
			end
		end
	end

	return false
end


-- Treat the table as a set, which means that duplicates are not allowed.
function table.insertSet (t, item)
	if (t) then
		if (not table.contains (t, item)) then
			table.insert (t, item)
		end
	end
end


function table.concatKeys (t, sep)
	local s = ""

	if (t) then
		for k in pairs (t) do
			s = s .. sep .. k
		end
		if (s ~= "") then
			s = s:sub (#sep + 1) -- Remove the leading separator.
		end
	end

	return s
end


local function padLeft (s, length, char)
	s = tostring (s)
	return char:rep (length - #s) .. s
end


-------------------------------------------------------------------------------
-------------------------------- Luup Utilities -------------------------------
-------------------------------------------------------------------------------


local function getPartitionState (partNo, variable)
	local device = g_partitions[partNo].device
	if (device) then
		local v = luup.variable_get (SID.PARTITION, variable, device)
		debug (string.format ("(getPartitionState) Got %s='%s' for partition %d (device #%d).", variable, tostring (v), partNo, device))
		if (not v or v == "") then
			log ("(getPartitionState) WARNING: Failed to get the value of '" .. variable .. "'.")
			return nil
		end
		return v
	else
		log (string.format ("(getPartitionState) ERROR: No device for partition %d (variable=%s).", partNo, variable))
		return nil
	end
end


local function setPartitionState (partNo, variable, value)
	local device = g_partitions[partNo].device
	if (device) then
		debug (string.format ("(setPartitionState) Set %s='%s' for partition %d (device #%d).", variable, tostring (value), partNo, device))
		luup.variable_set (SID.PARTITION, variable, value, device)
	else
		log (string.format ("(setPartitionState) ERROR: No device for partition %d (variable=%s, value='%s').", partNo, variable, tostring (value)))
	end
end


local function getPartitionNumber (id)
	if (type (id) == "number") then -- The 'id' is the device number.
		for k, v in pairs (g_partitions) do
			if (id == v.device) then
				return k
			end
		end
	else -- The 'id' is the device altid.
		local partNo = id:match ("vista_partition_(%d+)")
		return tonumber (partNo, 10)
	end

	return nil
end


-------------------------------------------------------------------------------
---------------------------- Action Implementations ---------------------------
-------------------------------------------------------------------------------


function requestArmMode (device, armMode, pinCode)

	debug (string.format ("(requestArmMode) Request new arm mode: %s, device #%d.", armMode, device))

	-- Check if the PIN code exists and is valid.
	if (not pinCode) then
		log ("(requestArmMode) ERROR: Cannot arm/disarm without PIN code.")
		task ("Cannot arm/disarm without PIN code.", TASK.ERROR)
		return false
	elseif (not pinCode:match ("^%d%d%d%d$")) then
		log ("(requestArmMode) ERROR: Invalid PIN code.")
		task ("Invalid PIN code.", TASK.ERROR)
		return false
	end

	-- Get the partition number.
	local partNo = (g_nPartitions > 1) and getPartitionNumber (device) or 1
	debug ("(requestArmMode) partNo=" .. partNo)

	debug ("(requestArmMode) Current arm mode: '" .. tostring (g_partitions[partNo].detailedArmMode) .. "'")

	if (g_nPartitions > 1) then -- Send the GOTO command first.
		local command = pinCode .. "*" .. partNo -- GOTO <partNo>
		luup.io.write (command)
		luup.sleep (SLEEP_MS)
	end

	-- Send the arming command.
	local command = pinCode .. armingCommands[armMode]
	luup.io.write (command)

	-- Store the PIN code to be used for sending '*'.
	g_partitions[partNo].pinCode = pinCode

	return true
end


function requestLetteredKeyPress (device, key, pinCode)

	debug (string.format ("(requestLetteredKeyPress) Request lettered key '%s' press, device #%d.", key, device))

	-- Check if the PIN code exists and is valid.
	if (not pinCode) then
		log ("(requestLetteredKeyPress) ERROR: Cannot use macro key without PIN code.")
		task ("Cannot use macro key without PIN code.", TASK.ERROR)
		return false
	elseif (not pinCode:match ("^%d%d%d%d$")) then
		log ("(requestLetteredKeyPress) ERROR: Invalid PIN code.")
		task ("Invalid PIN code.", TASK.ERROR)
		return false
	end

	-- Find the partition number.
	local partNo = (g_nPartitions > 1) and getPartitionNumber (device) or 1
	debug ("(requestLetteredKeyPress) partNo = " .. partNo)

	if (g_nPartitions > 1) then -- Send the GOTO command first.
		local command = pinCode .. "*" .. partNo -- GOTO
		luup.io.write (command)
		luup.sleep (SLEEP_MS)
	end

	-- Send the lettered key long press command.
	local command = keyCommands[key]
	g_userCodes[partNo] = pinCode
	luup.io.write (command)

	-- Store the PIN code to be used for sending '*'.
	g_partitions[partNo].pinCode = pinCode

	return true
end


function bypassZones (device, zones, pinCode)

	debug (string.format ("(bypassZones) Bypass zones: %s, device #%d.", tostring (zones), device))

	-- Check if the PIN code exists and is valid.
	if (not pinCode) then
		log ("(bypassZones) ERROR: PIN code required.")
		task ("PIN code required for bypassing zone(s).", TASK.ERROR)
		return false
	elseif (not pinCode:match ("^%d%d%d%d$")) then
		log ("(bypassZones) ERROR: Invalid PIN code.")
		task ("Invalid PIN code.", TASK.ERROR)
		return false
	end

	-- Find the partition number.
	local partNo = (g_nPartitions > 1) and getPartitionNumber (device) or 1
	debug ("(bypassZones) partNo = " .. partNo)

	if (g_nPartitions > 1) then -- Send the GOTO command first.
		local command = pinCode .. "*" .. partNo
		luup.io.write (command)
		luup.sleep (SLEEP_MS)
	end

	local command = ""
	if (zones and zones ~= "all") then -- Send the bypass command for each zone.
		for zone in zones:gmatch ("%d+") do
			command = pinCode .. "6" .. padLeft (zone, 2, "0")
			luup.io.write (command)
			luup.sleep (SLEEP_MS)
		end
	else -- Send the quick bypass command.
		command = pinCode .. "6#"
		luup.io.write (command)
	end

	-- Store the PIN code to be used for sending '*'.
	g_partitions[partNo].pinCode = pinCode

	return true
end


function toggleChimeMode (device, pinCode)

	debug ("(toggleChimeMode) Toggle chime mode for device #" .. device .. ".")

	-- Check if the PIN code exists and is valid.
	if (not pinCode) then
		log ("(toggleChimeMode) ERROR: PIN code required.")
		task ("PIN code required to change chime mode.", TASK.ERROR)
		return false
	elseif (not pinCode:match ("^%d%d%d%d$")) then
		log ("(toggleChimeMode) ERROR: Invalid PIN code.")
		task ("Invalid PIN code.", TASK.ERROR)
		return false
	end

	-- Find the partition number.
	local partNo = (g_nPartitions > 1) and getPartitionNumber (device) or 1
	debug ("(toggleChimeMode) partNo = " .. partNo)

	if (g_nPartitions > 1) then -- Send the GOTO command first.
		local command = pinCode .. "*" .. partNo -- GOTO
		luup.io.write (command)
		luup.sleep (SLEEP_MS)
	end

	-- Send the chime mode command.
	local command = pinCode .. "9"
	luup.io.write (command)

	-- Store the PIN code to be used for sending '*'.
	g_partitions[partNo].pinCode = pinCode

	return true
end


function setChimeMode (device, mode, pinCode)
	debug (string.format ("(setChimeMode) Set chime mode to '%s' for device #%s.", tostring (mode), device))

	-- Check if the PIN code exists and is valid.
	if (not pinCode) then
		log ("(setChimeMode) ERROR: PIN code required.")
		task ("PIN code required for changing the chime mode.", TASK.ERROR)
		return false
	elseif (not pinCode:match ("^%d%d%d%d$")) then
		log ("(setChimeMode) ERROR: Invalid PIN code.")
		task ("Invalid PIN code.", TASK.ERROR)
		return false
	end

	-- Find the partition number.
	local partNo = (g_nPartitions > 1) and getPartitionNumber (device) or 1
	debug ("(setChimeMode) partNo = " .. partNo)

	if (g_nPartitions > 1) then -- Send the GOTO command first.
		local command = pinCode .. "*" .. partNo
		luup.io.write (command)
		luup.sleep (SLEEP_MS)
	end

	if (g_partitions[partNo].chimeMode ~= mode) then -- We're not already in the requested mode.
		local command = pinCode .. "9"
		luup.io.write (command)
	end

	-- Store the PIN code to be used for sending '*'.
	g_partitions[partNo].pinCode = pinCode

	return true
end


function storePinCode (device, pinCode)

	debug ("(storePinCode) Store PIN code for device #" .. device .. ".")

	-- Check if the PIN code exists and is valid.
	if (not pinCode) then
		log ("(storePinCode) ERROR: PIN code missing.")
		task ("No code in the PIN code input.", TASK.ERROR)
		return false
	elseif (not pinCode:match ("^%d%d%d%d$")) then
		log ("(storePinCode) ERROR: Invalid PIN code.")
		task ("Invalid PIN code.", TASK.ERROR)
		return false
	end

	-- Find the partition number.
	local partNo = (g_nPartitions > 1) and getPartitionNumber (device) or 1
	debug ("(storePinCode) partNo = " .. partNo)

	-- Store the PIN code to be used for sending '*'.
	g_partitions[partNo].pinCode = pinCode

	return true
end


function requestPanicMode (device, state)
	log ("(requestPanicMode) ERROR: Action not implemented.")
	task ("RequestPanicMode not implemented.", TASK.ERROR)
	return false
end


-------------------------------------------------------------------------------
------------------------------ Message Processing -----------------------------
-------------------------------------------------------------------------------


local function addFaultedZone (zone)
	if (not zone) then
		log ("(addFaultedZone) ERROR: Invalid zone.")
		return
	end

	local newZone = (not g_faultedZones[zone]) and true or false -- Check if this zone isn't already in the list.

	g_faultedZones[zone] = TTL -- Add or refresh the TTL.
	debug (string.format ("(addFaultedZone) Zone %d TTL = %d", zone, g_faultedZones[zone]))

	if (newZone) then -- Update this variable only if we have a new zone.
		luup.variable_set (SID.PANEL, "FaultedZones", table.concatKeys (g_faultedZones, ","), lug_device)
		debug ("(addFaultedZone) Added zone #" .. zone .. ".")
	end
end


local function removeFaultedZone (zone)
	if (not zone) then
		log ("(removeFaultedZone) ERROR: Invalid zone.")
		return
	end

	if (not g_faultedZones[zone]) then
		debug (string.format ("(removeFaultedZone) Zone %d not in the faulted zones list.", zone))
		return
	end

	g_faultedZones[zone] = nil

	local faultedZones = table.concatKeys (g_faultedZones, ",")
	if (faultedZones == "") then
		faultedZones = "None"
	end
	luup.variable_set (SID.PANEL, "FaultedZones", faultedZones, lug_device)

	debug ("(removeFaultedZone) Removed zone #" .. zone .. ".")
end


function ttlCountdownTimer()
	for zone, ttl in pairs (g_faultedZones) do
		ttl = ttl - TIMER_INTERVAL
		if (ttl <= 0) then
			removeFaultedZone (zone)
		else
			g_faultedZones[zone] = ttl
			debug (string.format ("(ttlCountdownTimer) Zone %d TTL = %d", zone, ttl))
		end
	end

	luup.call_delay ("ttlCountdownTimer", TIMER_INTERVAL, "")
end


function sendStar (partNo)
	partNo = tonumber (partNo, 10)

	-- If we have more than 1 partition, send a GOTO command first.
	if (g_nPartitions > 1) then
		if (g_partitions[partNo].pinCode) then
			local command = g_partitions[partNo].pinCode .. "*" .. partNo -- GOTO <partNo>
			luup.io.write (command)
			luup.sleep (SLEEP_MS)
		else
			log ("(sendStar) No PIN code in memory for partition " .. partNo .. ".")
			return false
		end
	end

	debug ("(sendStar) Send '*' to partition " .. partNo .. ".")
	luup.io.write ("*")

	return true
end


local function updateArmMode (flags, partNo)

	-- Find the arm mode.
	if (not flags.ARMED_AWAY and not flags.ARMED_STAY) then -- Disarmed
		g_partitions[partNo].armMode = "Disarmed"
		if (flags.READY) then -- Ready
			g_partitions[partNo].detailedArmMode = "Ready"
		else -- Not ready
			g_partitions[partNo].detailedArmMode = "NotReady"
			luup.call_delay ("sendStar", 1, tostring (partNo)) -- Send '*' to get the faulted zones.
		end
	elseif (flags.ARMED_AWAY) then
		if (not flags.BEEP) then
			g_partitions[partNo].armMode = "Armed"
			if (not flags.ENTRY_DELAY_OFF) then -- Armed AWAY
				g_partitions[partNo].detailedArmMode = "Armed"
			else -- Armed MAXIMUM
				g_partitions[partNo].detailedArmMode = "ArmedInstant"
			end
		else -- Exit delay
			g_partitions[partNo].detailedArmMode = "ExitDelay"
		end
	else
		if (not flags.BEEP) then
			g_partitions[partNo].armMode = "Armed"
			if (not flags.ENTRY_DELAY_OFF and not flags.PERIMETER_ONLY) then -- Armed STAY
				g_partitions[partNo].detailedArmMode = "Stay"
			elseif (flags.ENTRY_DELAY_OFF) then -- Armed INSTANT
				g_partitions[partNo].detailedArmMode = "StayInstant"
			else -- Armed NIGHT-STAY
				g_partitions[partNo].detailedArmMode = "Night"
			end
		else -- Exit delay
			g_partitions[partNo].detailedArmMode = "ExitDelay"
		end
	end

	local currentArmMode = getPartitionState (partNo, "ArmMode") or "nil"
	if (g_partitions[partNo].armMode and g_partitions[partNo].armMode ~= currentArmMode) then
		debug (string.format ("(getArmMode) current ArmMode = '%s', new ArmMode = '%s'", currentArmMode, g_partitions[partNo].armMode))
		setPartitionState (partNo, "ArmMode", g_partitions[partNo].armMode)
	end

	local currentDetailedArmMode = getPartitionState (partNo, "DetailedArmMode") or "nil"
	if (g_partitions[partNo].detailedArmMode and g_partitions[partNo].detailedArmMode ~= currentDetailedArmMode) then
		debug (string.format ("(getArmMode) current DetailedArmMode = '%s', new DetailedArmMode = '%s'", currentDetailedArmMode, g_partitions[partNo].detailedArmMode))
		setPartitionState (partNo, "DetailedArmMode", g_partitions[partNo].detailedArmMode)
	end

	return true
end


local function updateChimeMode (flags, partNo)
	local chimeMode = flags.CHIME_MODE and "1" or "0"
	if (g_partitions[partNo].chimeMode ~= chimeMode) then
		g_partitions[partNo].chimeMode = chimeMode
		luup.variable_set (SID.PANEL, "ChimeMode", chimeMode, g_partitions[partNo].device)
	end
end


local function checkForAlarms (flags, partNo)

	if (flags.FIRE_ALARM or flags.ALARM) then
		local alarmText = flags.FIRE_ALARM and "FIRE ALARM!" or "ALARM!"
		log ("(checkForAlarms) " .. alarmText)
		task (alarmText, TASK.ERROR_PERM)

		local alarm = getPartitionState (partNo, "Alarm")
		if (alarm ~= "Active") then
			setPartitionState (partNo, "Alarm", "Active")
		end

		local alarmMemory = flags.ALARM_MEMORY and "1" or "0"
		setPartitionState (partNo, "AlarmMemory", alarmMemory)

		return true
	else
		local alarm = getPartitionState (partNo, "Alarm")
		if (alarm ~= "None") then
			setPartitionState (partNo, "Alarm", "None")
		end
	end
end


local function getAddressBitMask (binaryData)

	-- Bytes 2-5 contain the keypad addresses.
	local addressMask = binaryData:match ("^%x%x(%x%x%x%x%x%x%x%x).*")
	addressMask = tonumber (addressMask, 16)
	debug ("(getAddressBitMask) addressMask = " .. addressMask)

	local bitMask = {}
	for i = 1, 32 do
		table.insert (bitMask, addressMask % 2)
		addressMask = math.floor (addressMask / 2)
	end

	local debugString = "(getAddressBitMask) bitMask = " ..
	                    table.concat (bitMask, "", 1, 8) .. " " ..
	                    table.concat (bitMask, "", 9, 16) .. " " ..
	                    table.concat (bitMask, "", 17, 24) .. " " ..
	                    table.concat (bitMask, "", 25, 32)
	debug (debugString)
	return bitMask
end


local function getTargetPartitions (binaryData)

	local addressBits = getAddressBitMask (binaryData)
	local targetPartitions = {}
	for addr, part in pairs (g_addrPart) do
		if (addressBits[ADD_IDX[addr]] == 1) then
			table.insertSet (targetPartitions, part)
		end
	end

	debug ("(getTargetPartitions) Target partitions: " .. table.concat (targetPartitions, ","))
	return targetPartitions
end


local function getStatusFlags (bitMask)
	local flags = {}

	flags.READY = (bitMask:match ("%d", 1) == "1") and true or false
	flags.ARMED_AWAY = (bitMask:match ("%d", 2) == "1") and true or false
	flags.ARMED_STAY = (bitMask:match ("%d", 3) == "1") and true or false
	flags.PROGRAMMING_MODE = (bitMask:match ("%d", 5) == "1") and true or false
	flags.BEEP = (tonumber (bitMask:match ("%d", 6), 10) >= 1) and true or false
	flags.BYPASSED = (bitMask:match ("%d", 7) == "1") and true or false
	flags.CHIME_MODE = (bitMask:match ("%d", 9) == "1") and true or false
	flags.ALARM_MEMORY = (bitMask:match ("%d", 10) == "1") and true or false
	flags.ALARM = (bitMask:match ("%d", 11) == "1") and true or false
	flags.ENTRY_DELAY_OFF = (bitMask:match ("%d", 13) == "1") and true or false
	flags.FIRE_ALARM = (bitMask:match ("%d", 14) == "1") and true or false
	flags.PERIMETER_ONLY = (bitMask:match ("%d", 16) == "1") and true or false

	-- Create variable to use for debugging.
	local fdebug = ""
	for k, v in pairs (flags) do
		if (v) then
			fdebug = fdebug .. ", " .. tostring (k)
		end
	end
	if (fdebug ~= "") then
		fdebug = fdebug:sub (3) -- Remove the first comma and the first space.
	end
	debug ("(getStatusFlags) Active flags: " .. fdebug)

	return flags
end


function processIncoming (data)

	if (not g_nPartitions) then
		log ("(processIncoming) ERROR: NumPartitions not set. Terminate.")
		return false
	elseif (g_noAddrPart) then
		log ("(processIncoming) ERROR: No keypads for one or more partitions. Terminate.")
		return false
	end

	if (not data or data == "") then
		return false
	elseif (data:sub (1, 1) == '!') then
		debug ("(processIncoming) Drop message.")
		return true
	end

	-- Split the message into sections.
	local sections = {}
	sections[1], sections[2], sections[3], sections[4] = data:match ('%[([%d%-]+)%],(%x+),%[(%x+)%],"(.+)"')

	-- Log each section and check for validity.
	local invalidMessage = false
	for i = 1, 4 do
		debug (string.format ("(processIncoming) Section %d: %s", i, tostring (sections[i])))
		if (not sections[i]) then
			invalidMessage = true
		end
	end
	if (invalidMessage) then
		log ("(processIncoming) ERROR: Invalid message.")
		return false
	end

	local flags = getStatusFlags (sections[1])
	if (flags.PROGRAMMING_MODE) then
		debug ("(processIncoming) Programming mode, drop message.")
		return true
	end

	--
	-- Process message.
	--
	-- The FAULT and BYPAS messages aren't related to any partition.
	if (sections[4]:match ("^FAULT %d%d")) then -- Zone faulted
		addFaultedZone (tonumber (sections[2], 10))
	elseif (sections[4]:match ("^BYPAS %d%d")) then -- Zone bypassed
		removeFaultedZone (tonumber (sections[2], 10))
	else
		-- Get the partitions this message is for.
		local targetPartitions = {1}
		if (g_nPartitions > 1) then
			targetPartitions = getTargetPartitions (sections[3])
		end

		for _, partNo in pairs (targetPartitions) do
			-- Check if we have any alarms.
			if (checkForAlarms (flags, partNo)) then
				return false
			end

			-- Update the chime mode.
			updateChimeMode (flags, partNo)

			if (sections[4]:lower():match ("enter user code")) then -- A lettered key was pressed.
				debug ("(processIncoming) A lettered key was pressed, send user code...")
				luup.io.write (tostring (g_userCodes[partNo]))
				g_userCodes[partNo] = nil
			else -- Chances are this message is an alarm status report.
				updateArmMode (flags, partNo)
			end
		end
	end

	return true
end


-------------------------------------------------------------------------------
--------------------------- Initialization Functions --------------------------
-------------------------------------------------------------------------------


local function getPartitionDevices()
	for k, v in pairs (luup.devices) do
		if (v.device_num_parent == lug_device) then
			local partNo = getPartitionNumber (v.id)
			if (partNo) then
				debug (string.format ("(getPartitionDevices) Partition %d device number = %d", partNo, k))
				g_partitions[partNo] = {device = k}
			else
				log (string.format ("(getPartitionDevices) Found unexpected child: device #%d, id=%s", k, tostring (v.id)))
				task (string.format ("Found unexpected child: device #%d, id=%s", k, tostring (v.id)), TASK.ERROR)
			end
		end
	end
end


local function appendPartitions (rootPtr)
	for i = 1, g_nPartitions do
		debug (string.format ("(appendPartitions) Append partition %d.", i))
		local parameters = SID.PANEL .. ",KeypadAddresses="
		luup.chdev.append (lug_device, rootPtr, string.format ("vista_partition_%d", i), string.format ("Ademco Vista Partition %d", i), "urn:schemas-micasaverde-com:device:AlarmPartition:2", "D_VistaAlarmPartition1.xml", "", parameters, false)
	end
end


local function getKeypadPartitions()
	-- 'KeypadAddresses' is comma separated value which contains
	-- the addresses of the keypads assigned to each partition.
	-- e.g KeypadAddresses = '16,18', the keypads at addresses 16 and 18 are assigned to this partition.

	g_noAddrPart = false
	for part, v in  pairs (g_partitions) do
		local keypadAddresses = luup.variable_get (SID.PANEL, "KeypadAddresses", v.device)
		debug (string.format ("(getKeypadPartitions) Got addresses: %s for partition %d.", tostring (keypadAddresses), part))
		if (not keypadAddresses or keypadAddresses == "") then
			g_noAddrPart = true
			log ("(getKeypadPartitions) ERROR: No keypad for partition " .. part .. ". Terminate.")
			return false, "No keypad assigned to partition " .. part .. "."
		else
			keypadAddresses = keypadAddresses:split (",")
			for _, addr in pairs (keypadAddresses) do
				g_addrPart[padLeft (addr, 2, "0")] = part
			end
		end
	end

	return true
end


local function getNumPartitions()
	local nPartitions = luup.variable_get (SID.PANEL, "NumPartitions", lug_device)
	debug ("(getNumPartitions) nPartitions = " .. tostring (nPartitions))
	if (not nPartitions or nPartitions == "0") then
		if (not nPartitions) then
			luup.variable_set (SID.PANEL, "NumPartitions", "0", lug_device)
		end
		log ("(getNumPartitions) ERROR: NumPartitions not set. Terminate.")
		return false, "'NumPartitions' not set."
	else
		g_nPartitions = tonumber (nPartitions, 10)
		if (not g_nPartitions) then
			log ("(getNumPartitions) ERROR: Invalid NumPartitions value. Terminate.")
			return false, "Invalid 'NumPartitions' value."
		end
	end

	return true
end


-------------------------------------------------------------------------------
----------------------------------- Main --------------------------------------
-------------------------------------------------------------------------------


function main (lul_device)
	log ("Honeywell Ademco Vista Panel Plugin version " .. VERSION)

	lug_device = lul_device
	luup.variable_set (SID.PANEL, "PluginVersion", VERSION, lug_device)

	-- Check if Debug Mode is enabled.
	local debugMode = luup.variable_get (SID.PANEL, "DebugMode", lug_device)
	if (debugMode == nil) then
		luup.variable_set (SID.PANEL, "DebugMode", DEBUG_MODE_DEFAULT, lug_device)
		debugMode = DEBUG_MODE_DEFAULT
	end
	DEBUG_MODE = (debugMode == "1") and true or false

	-- Get the number of partitions.
	local status, message = getNumPartitions()
	if (not status) then
		return false, message, "Ademco Vista Plugin"
	end

	-- Create the partitions.
	local rootPtr = luup.chdev.start (lug_device)
	appendPartitions (rootPtr)
	luup.chdev.sync (lug_device, rootPtr)

	getPartitionDevices()

	-- Get the partitions assigned to each keypad.
	status, message = getKeypadPartitions()
	if (not status) then
		return false, message, "Ademco Vista Plugin"
	end

	luup.variable_set (SID.PANEL, "FaultedZones", "None", lug_device)

	-- Get the TTL.
	local ttl = luup.variable_get (SID.PANEL, "TTL", lug_device)
	if (not ttl) then
		luup.variable_set (SID.PANEL, "TTL", TTL, lug_device)
	else
		TTL = tonumber (ttl, 10) or TTL
	end

	-- Start the timer.
	ttlCountdownTimer()

	debug ("(Vista Plugin Startup) SUCCESS: Startup successful.")
	return true, "Startup successful.", "Vista Alarm Panel"
end
